/********************************************************
 *  ██████╗  ██████╗████████╗██╗
 * ██╔════╝ ██╔════╝╚══██╔══╝██║
 * ██║  ███╗██║        ██║   ██║
 * ██║   ██║██║        ██║   ██║
 * ╚██████╔╝╚██████╗   ██║   ███████╗
 *  ╚═════╝  ╚═════╝   ╚═╝   ╚══════╝
 * Geophysical Computational Tools & Library (GCTL)
 *
 * Copyright (c) 2022  Yi Zhang (yizhang-geo@zju.edu.cn)
 *
 * GCTL is distributed under a dual licensing scheme. You can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either version 2 
 * of the License, or (at your option) any later version. You should have 
 * received a copy of the GNU Lesser General Public License along with this 
 * program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * If the terms and conditions of the LGPL v.2. would prevent you from using 
 * the GCTL, please consider the option to obtain a commercial license for a 
 * fee. These licenses are offered by the GCTL's original author. As a rule, 
 * licenses are provided "as-is", unlimited in time for a one time fee. Please 
 * send corresponding requests to: yizhang-geo@zju.edu.cn. Please do not forget 
 * to include some description of your company and the realm of its activities. 
 * Also add information on how to contact you by electronic and paper mail.
 ******************************************************/

#ifndef _GCTL_GRAV_KERNEL_TETRAHEDRON_H
#define _GCTL_GRAV_KERNEL_TETRAHEDRON_H

#include "gm_data.h"

namespace gctl
{
	struct gravtet_para
	{
		tensor F[4]; ///< tensor products of four facets
		tensor E[12]; ///< tensor products of six edges
		double edglen[12]; ///< edge lengths of six edges
	};

	typedef type_tetrahedron<gravtet_para> grav_tetrahedron; ///< 带gravtet_para属性的四面体结构体

	/**
	 * @brief      calculate the gravity parameters of given element type.
	 *
	 * @param      in_tri    Input elements
	 * @param      out_para  The output parameter array
	 */
	void callink_gravity_para(array<grav_tetrahedron> &in_tet, array<gravtet_para> &out_para);

    /**
	 * @brief      计算直角坐标下的三维四面体单元体正演核矩阵
	 *
	 * @param      out_kernel     输出的核矩阵
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gkernel(matrix<double> &out_kernel, const array<grav_tetrahedron> &ele, const array<point3dc> &obsp, 
		gravitational_field_type_e comp_id = Vz, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      计算球坐标下的三维四面体单元体正演核矩阵
	 *
	 * @param      out_kernel     输出的核矩阵
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           球坐标下的观测点数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gkernel(matrix<double> &out_kernel, const array<grav_tetrahedron> &ele, const array<point3ds> &obsp, 
		gravitational_field_type_e comp_id = Vz, verbose_type_e verbose = FullMsg);

    /**
	 * @brief      计算直角坐标下的三维四面体单元体重力观测值正演
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<double> &out_obs, const array<grav_tetrahedron> &ele, const array<point3dc> &obsp, 
		const array<double> &rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      计算直角坐标下的三维四面体单元体重力观测值正演
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           三维球坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<point3dc> &out_obs, const array<grav_tetrahedron> &ele, const array<point3dc> &obsp, 
		const array<double> &rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      计算直角坐标下的三维四面体单元体重力观测值正演
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           三维球坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<tensor> &out_obs, const array<grav_tetrahedron> &ele, const array<point3dc> &obsp, 
		const array<double> &rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      计算球坐标下的三维四面体单元体重力观测值正演
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<double> &out_obs, const array<grav_tetrahedron> &ele, const array<point3ds> &obsp, 
		const array<double> &rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      计算球坐标下的三维四面体单元体重力观测值正演
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维四面体单元体数组
	 * @param[in]  obsp           三维球坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<point3dc> &out_obs, const array<grav_tetrahedron> &ele, const array<point3ds> &obsp, 
		const array<double> &rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      计算球坐标下的三棱锥单元体正演重力数据
	 * 
	 * @note       三棱锥类型仅用于球坐标下的重力数据正演
	 *
	 * @param      out_obs        输出的重力数据
	 * @param[in]  ele            三棱锥单元体数组
	 * @param[in]  obsp           球坐标下的观测点数组
	 * @param[in]  rho            单元体密度数组
	 * @param[in]  comp_id        待计算的重力场分量 默认为重力值
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<tensor> &out_obs, const array<grav_tetrahedron> &ele, const array<point3ds> &obsp, 
		const array<double> &rho, verbose_type_e verbose = FullMsg);
}

#endif // _GCTL_GRAV_KERNEL_TETRAHEDRON_H